// arq module arq.cxx
// Copyright (c) 2007, Dave Freese, W1HKJ
//
//
// arq.cxx / arq.h is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// arq.cxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with tUr program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "arq.h"

#include <sys/time.h>

using namespace std;

string arq::upcase(string s)
{
	for (size_t i = 0; i < s.length(); i++)
		s[i] = toupper(s[i]);
	return s;
}

arq::arq()
{
	sendfnc = NULL;
	rcvfnc = NULL;
	putfnc = NULL;
	unprotofnc = NULL;
	qualityfnc = NULL;
	
	Header.erase();

	MyStreamID = '0';
	UrStreamID = '0';
	
	UrCall.erase();
	MyCall.erase();

	logfile = "server.log";

// queues //
	TxTextQueue.erase();		// Text out to mail engine
	RxTextQueue.erase();		// Text in from mail engine
	TxPlainTextQueue.erase();
	RxFrameQueue.erase();
	lastRxChar = 0;
	TXflag = false;

	Sessionnumber = 0;
	
	exponent = EXPONENT;
	maxheaders = MAXHEADERS;
	RetryTime = RETRYTIME;
	Retries = RETRIES;
	Timeout = TIMEOUT;
	TxDelay = TXDELAY;

	setBufferlength();
	

// status variables //
	totalRx = 0;
	nbrbadRx = 0;
	totalTx = 0;
	nbrbadTx = 0;
	payloadlength = 32;		// Average length of payload received

// static status
	Firstsent = MAXCOUNT - 1; 		// First Header  I sent last turn
	LastHeader = MAXCOUNT - 1;		// Last Header I sent last turn
	Lastqueued = MAXCOUNT - 1;		// Last Header in static send queue

	EndHeader = MAXCOUNT - 1;		// Last  I received o.k.
	GoodHeader = MAXCOUNT - 1;		// Last Header received consecutive

// Ur status
	UrGoodHeader = MAXCOUNT - 1;	// Other station's Good Header
	UrLastHeader = MAXCOUNT - 1;	// Other station's Header last sent
	UrEndHeader = MAXCOUNT - 1;		// Other station's last received Header
	blkcount = -1;
	
	TXflag = false;					// TX on
	LinkState = DOWN;				// ARQ link is initially down
	Sending = 0;
	PollOK = false;

	MyMissing.clear();
	MissingRxBlocks = "";
	
	TxBlocks.clear();
	TxMissing.clear();

	RxPending.clear();

 	arqstop = false;
 	
 	retries = Retries;
 	retrytime = 0; // RetryTime / ARQLOOPTIME;
 	timeout = Timeout / ARQLOOPTIME;
 	
 	rx2txdelay = TxDelay / ARQLOOPTIME;
}

void arq::setBufferlength()
{
	Bufferlength = 1;
	for (int i = 0; i < exponent; i++) Bufferlength *= 2;
	MyBlockLengthChar = '0' + exponent;
}

void arq::resetTx()
{
	Firstsent = MAXCOUNT - 1; 		// First Header  I sent last turn
	LastHeader = MAXCOUNT - 1;		// Last Header I sent last turn
	Lastqueued = MAXCOUNT - 1;		// Last Header in static send queue
	TxMissing.clear();
	TxBlocks.clear();
	TxTextQueue.clear();
//	UrMissing.clear();
}

void arq::resetRx()
{
	RxTextQueue.erase();		// Text in from mail engine
	RxFrameQueue.erase();
	lastRxChar = 0;
	EndHeader = MAXCOUNT - 1;		// Last  I received o.k.
	GoodHeader = MAXCOUNT - 1;		// Last Header I received conseq. o.k, 1st in send queue
	RxPending.clear();
}

void arq::reset()
{
	resetTx();
	resetRx();
	blkcount = -1;
}

// new session number
// unknown stream id = 0
// known id's from 1 to 63
void arq::newsession() 
{
	if (++SessionNumber == 64) SessionNumber = 1;
	MyStreamID = SessionNumber + '0';
}

// get new blocknumber
void arq::newblocknumber() 
{
	Lastqueued++;
	Lastqueued %= MAXCOUNT;
}

// Checksum of header + Header
string arq::checksum(string &s)
{
	framecrc.reset();
	return framecrc.scrc16(s);
}

// Start header when MyStreamID has been assigned
void arq::newHeader()
{
	Header.erase();
	Header += SOH;
	Header += '0'; // protocol version;
}

void arq::IdHeader()
{
	newHeader();
	Header += UrStreamID;
}

void arq::UnkHeader()
{
	newHeader();
	Header += '0';
}

void arq::addToTxQue(string s)
{
	TxTextQueue += '\n';
	TxTextQueue.append(s);
	totalTx++;
}

// Connect (client:port, server:port)
// c Header = Client:port Server:port <streamnr. (0)> <max. Headerlen>
// e.g.: '00cW1HKJ:1025 KH6TY:24 4'
//
void arq::connectFrame() 
{
	reset();
	UnkHeader();
	Header += CONREQ;

	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":1025");
	Payload += ' ';
	Payload.append(UrCall);
	Payload.append(":24");
	Payload += ' ';
	Payload += MyStreamID;
	Payload += ' ';
	Payload += MyBlockLengthChar;

	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);

	LinkState = CONNECTING;
} 

// Connect acknowledge (server:port, client:port)
// k Header = Server:port Client:port <streamnr.> <max. Headerlen>
// e.g: '00kKH6TY:24 W1HKJ:1024 8 6'
//
void arq::ackFrame ()
{
	reset();
	IdHeader();
	Header += CONACK;
	
	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":24");
	Payload += ' ';
	Payload.append(UrCall);
	Payload += ' ';
	Payload += MyStreamID;
	Payload += ' ';
	Payload += MyBlockLengthChar;
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

// Connect (caller:port, static:port)
// c Header = Caller:port static:port <streamnr. (0)> <max. Headerlen>
// e.g.: '00cW1HKJ:87 KH6TY:87 4'
//
void arq::ttyconnectFrame() 
{
	UnkHeader();
	Header += CONREQ;
	
	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":87");
	Payload += ' ';
	Payload.append(UrCall);
	Payload.append(":87");
	Payload += ' ';
	Payload += MyStreamID;
	Payload += ' ';
	Payload += MyBlockLengthChar;
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
} 

// Connect acknowledge (server:port, client:port)
// k Header = Server:port Client:port <streamnr.> <max. Headerlen>
// e.g: '00kKH6TY:87 W1HKJ 4'
// Service id # 87 is keyboard-to-keyboard
//
void arq::ttyackFrame() 
{
	IdHeader();
	Header += CONACK;
	
	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":87");
	Payload += ' ';
	Payload.append(UrCall);
	Payload += ' ';
	Payload += MyBlockLengthChar;
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

// Identify
//i frame = '00iKH6TY de W1HKJ'
void arq::identFrame()
{
	IdHeader();
	Header += IDENT;
	
	Payload.erase();
	Payload.append(UrCall);
	Payload.append(" de ");
	Payload.append(MyCall);
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

// e.g. Ping frame
// u Header = From:port
// e.g: '00uKH6TY:7 '
void arq::pingFrame()
{
	IdHeader();
	Header += UNPROTO;
	
	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":7");
	Payload += ' ';
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;
	
	addToTxQue(Frame);
}

// Status report (End, Good, Lastrx, Missing)
//p frame = <last Header tx><last Header rx ok><last Header rx> <missing Headers>
//e.g.: '00sXHCAB'
//
void arq::statFrame() {

	IdHeader();
	Header += STATUS;

	Payload.erase();
	Payload += (LastHeader + 0x20);
	Payload += (GoodHeader + 0x20);
	Payload += (EndHeader + 0x20);
	Payload.append(MissingRxBlocks);
	
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;
	
	addToTxQue(Frame);
}

// Disconnect session
//d frame = ""
//e.g.: '00d'

void arq::disconnectFrame()
{
	IdHeader();
	Payload.erase();
	Payload += DISREQ;
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

void arq::disackFrame()
{
	IdHeader();
	Payload.erase();
	Payload += DISACK;
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;
	addToTxQue(Frame);
}

// ABORT session
//a frame = ""
//e.g.: '00a'
void arq::abortFrame()
{
	IdHeader();
	Payload.erase();
	Payload += ABORT;
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

// Beacon frame
// u Header = From:port  data
// e.g: '00uKH6TY:72 Beacon text '
//
void arq::beaconFrame(string txt) 
{
	UnkHeader();
	Header += UNPROTO;

	Payload.erase();
	Payload.append(MyCall);
	Payload.append(":72");
	Payload += ' ';
	if (txt.length() > (size_t)Bufferlength)
		Payload.append(txt.substr(0, Bufferlength));
	else
		Payload.append(txt);
	Frame = Header + Payload;
	Frame = Frame + checksum(Frame);
	Frame += EOT;

	addToTxQue(Frame);
}

// poll
//p frame = <last Header tx><last Header rx ok><last Header rx> <missing Headers>
//e.g.: '00pXHCAB'
void arq::pollFrame() 
{

	IdHeader();
	Frame = Header;
	Frame += POLL;
	Frame += (LastHeader + 0x20);
	Frame += (GoodHeader + 0x20);
	Frame += (EndHeader + 0x20);
	Frame.append(MissingRxBlocks);
	Frame.append(checksum(Frame));
	Frame += EOT;

	addToTxQue(Frame);
}

// Text frame
void arq::textFrame(cTxtBlk block)
{
	IdHeader();
	Frame = Header;
	Frame += (block.nbr() + 0x20);
	Frame.append(block.text());
	Frame.append(checksum(Frame));
	Frame += SOH;

	addToTxQue(Frame);
}



//=====================================================================

void arq::parseIDENT()
{
}

void arq::parseCONREQ()
{

	size_t p1 = 0, p2 = rcvPayload.find(':');
	if (p2 == string::npos)
		return;
	if (LinkState == CONNECTED || LinkState == WAITFORACK) return; // disallow multiple connects
	
// requesting stations callsign
	UrCall = upcase(rcvPayload.substr(p1, p2 - p1));
	p1 = rcvPayload.find(' ', p2+1);
	if (p1 == string::npos) {
		UrCall.erase();
		return;
	}

	p1++;
	p2 = rcvPayload.find(":", p1);
	string testcall = upcase(rcvPayload.substr(p1, p2 - p1));
	if (testcall != MyCall) {
		UrCall.erase();
		return;
	}

	p1 = rcvPayload.find(' ', p2 +1);
	if (p1 == string::npos) {
		UrCall.erase();
//		std::cout << std::endl;
		return;
	}

	p1++; // *p1 ==> StreamID for requesting station
	UrStreamID = rcvPayload[p1];
	p1++; // *p1 ==> requested block size
	UrBlockLengthChar = rcvPayload[p1];

	reset();

	LinkState = WAITFORACK;
	newsession();

	TxTextQueue.erase();
	ackFrame();

}

void arq::parseCONACK()
{
	if (LinkState != CONNECTING) {
		return; // Connect Acknowledge only valid during a connect
	}
	
	size_t p1 = 0, p2 = rcvPayload.find(':');
//	LinkState = DOWN;
	if (p2 == string::npos)
		return;
// requesting stations callsign
	UrCall = upcase(rcvPayload.substr(p1, p2 - p1));
	p1 = rcvPayload.find(' ', p2+1);
	if (p1 == string::npos) {
		UrCall.erase();
		return;
	}

	p1++;
	p2 = rcvPayload.find(" ", p1);
	string testcall = upcase(rcvPayload.substr(p1, p2 - p1));
	if (testcall != MyCall) {
		UrCall.erase();
		return;
	}
	
	p1++; // *p1 ==> StreamID for requesting station
	UrStreamID = rcvPayload[p1];
	p1++; // *p1 ==> requested block size
	UrBlockLengthChar = rcvPayload[p1];

	RxTextQueue.erase();

	LinkState = CONNECTED;
	timeout = Timeout / ARQLOOPTIME;
	
	statFrame();
}

void arq::parseREFUSED()
{
}

void arq::parseDISREQ()
{
	TxTextQueue.erase();
	disackFrame();
	LinkState = DOWN;
}

void arq::parseDISACK()
{
	if (LinkState != DISCONNECTING) return;
	LinkState = DOWN;
}

void arq::parseABORT()
{
	reset();
	abortfnc();
	statFrame();
}

void arq::parseUNPROTO()
{
	size_t p1 = 0, p2 = rcvPayload.find(':');
	if (p2 == string::npos)
		return;
// requesting stations callsign
	UrCall = upcase(rcvPayload.substr(p1, p2 - p1));
	if (unprotofnc) unprotofnc(UrCall);
//	if (putfnc) putfnc(rcvPayload + '\n');
	if (putfnc) putfnc(rcvPayload + "\n");
}

char *szState[] = {
	"DOWN", "CONNECTED", "TIMEDOUT", "WAITING", "CONNECTING", "DISCONNECTING" };

void arq::parseSTATUS()
{
	UrLastHeader = rcvPayload[0] - 0x20;	// Other station's Header last sent
	UrGoodHeader = rcvPayload[1] - 0x20;	// Other station's Good Header
	UrEndHeader = rcvPayload[2] - 0x20;		// Other station's last received Header

	size_t nummissing = rcvPayload.length() - 3;
	vector<int> missing;
// create the missing list
// all reported missing blocks
	if (nummissing > 0)
		for (size_t i = 0; i < nummissing; i++)
			missing.push_back(rcvPayload[i+3] - 0x20);
// append those not reported missing from UrLastGood up to &
// including LastHeader
	if (UrEndHeader != LastHeader) {
		int m = UrEndHeader + 1;
		while (m != LastHeader) {
			missing.push_back(m);
			m++;
			if (m > MAXCOUNT) m -= MAXCOUNT;
		}
		missing.push_back(LastHeader);
	}
	
	if (missing.empty())
		TxMissing.clear();		
	else if (TxMissing.empty() == false) {
		list<cTxtBlk> keep;
		list<cTxtBlk>::iterator p = TxMissing.begin();
		while (p != TxMissing.end()) {
			for (size_t n = 0; n < missing.size(); n++) {
				if (p->nbr() == missing[n]) {
					keep.push_back(*p);
					break;
				}
			}
			p++;
		}
		TxMissing = keep;
	}
	

	switch (LinkState) {
		case WAITFORACK :
			LinkState = CONNECTED;
			break;
		case DISCONNECTING :
			LinkState = DOWN;
			break;
		case WAITING :
			LinkState = CONNECTED;
			break;
		case ABORTING :
			LinkState = CONNECTED;
			break;
		default: break;
	}

}

void arq::parsePOLL()
{
	if (LinkState == DOWN) return;
	statFrame();
	LinkState = CONNECTED;
}

void arq::parseDATA()
{
	vector<cTxtBlk>::iterator p1, p2;
	int n1, n2;
	
	if (LinkState < CONNECTED) return; // do not respond if DOWN or TIMEDOUT
	
	for (p1 = RxPending.begin(); p1 < RxPending.end(); p1++) 
		if (blknbr == p1->nbr()) {
			return;
		}
			
	cTxtBlk tempblk(blknbr, rcvPayload);
	RxPending.push_back (tempblk);

	for (p1 = RxPending.begin(); p1 < RxPending.end() - 1; p1++) {
		n1 = p1->nbr();
		if (n1 < GoodHeader) n1 += MAXCOUNT;
		for (p2 = p1 + 1; p2 < RxPending.end(); p2++) {
			n2 = p2->nbr();
			if (n2 < GoodHeader) n2 += MAXCOUNT;
			if (n2 < n1) {
				tempblk = *p1;
				*p1 = *p2;
				*p2 = tempblk;
			}
		}
	}
// compute new EndHeader
	EndHeader = GoodHeader;
	if (!RxPending.empty()) {
		p1 = RxPending.end() - 1;
		EndHeader = p1->nbr();
	}

// add RxPending blocks that are consecutive to GoodHeader
	p1 = RxPending.begin();	
	while (!RxPending.empty()) {
		if ((p1->nbr() != (GoodHeader +1) % MAXCOUNT))
			break;
		RxTextQueue.append(p1->text());
		GoodHeader = p1->nbr();
		if (putfnc) putfnc(p1->text());
		RxPending.erase(p1);
		p1 = RxPending.begin();
	}

	MissingRxBlocks = "";

	if (RxPending.empty())
		return;

	int start = (GoodHeader + 1)%MAXCOUNT;
	int end = (EndHeader + 1)%MAXCOUNT;
	int test;
	bool ok;
	if (end < start) end += MAXCOUNT;
	for (int i = start; i < end; i++) {
		test = (i % MAXCOUNT);
		ok = false;
		for (p1 = RxPending.begin(); p1 < RxPending.end(); p1++) {
			if (test == p1->nbr()) {
				ok = true;
				break;
			}
		}
		if (!ok)
			MissingRxBlocks += test + 0x20;
	}
}



// expects to receive a full frame
// txt[0] == SOH
// txt[len - 3] ... txt[len] == CRC
// returns
//   -1  invalid frame
//   -n  failed CRC for text type n
//    n  valid frame
//       rcvPayload will contain the valid payload
//
int arq::parseFrame(string txt)
{
	if ( txt.length() < 8 ) return -1; // not a valid frame
	Ccrc16 testcrc;
	size_t len = txt.length();

	totalRx++;
	
	string sRcvdCRC = testcrc.scrc16( txt.substr(0, len - 4));
	if (sRcvdCRC != txt.substr(len - 4) ) {
		nbrbadRx++;
		return -1; // failed CRC test
	}

 	retries = Retries;

// reset the retrytime timer
	retrytime = 0;
	
	if (LinkState == CONNECTED)
	 	timeout = Timeout / ARQLOOPTIME;

	rcvPayload = txt.substr(4, len - 8);
	fID = txt[3];
	blknbr = fID - 0x20;

	switch (fID) {
		case IDENT : 
			parseIDENT();
			break;
		case CONREQ : 
			parseCONREQ();
			break;
		case CONACK :
			parseCONACK();
			break;
		case REFUSED :
			parseREFUSED();
			break;
		case DISREQ :
			parseDISREQ();
			break;
		case DISACK :
			parseDISACK();
			break;
		case STATUS :
			parseSTATUS();
			break;
		case POLL :
			parsePOLL();
			break;
		case ABORT :
			parseABORT();
			break;
		case UNPROTO :
			parseUNPROTO();
			break;
		default :
			parseDATA();
	}
	return fID;
}


void arq::rcvChar( char c )
{
	if ( c == 0x06 ) {
		Sending = 0;
		return;
	}
	
	if (lastRxChar == SOH && c == SOH)
		return;
		
	if (lastRxChar == EOT && c == EOT)
		return;

	if ((c == SOH || c == EOT) && RxFrameQueue.empty() == false) {
		parseFrame(RxFrameQueue);
		RxFrameQueue.erase();
		if (c == SOH)
			RxFrameQueue += c;
	} else
		RxFrameQueue += c;
	lastRxChar = c;
}

//=====================================================================

void arq::sendText (string txt)
{
	size_t offset = 0;
	cTxtBlk tempblk;
	Blocks2Send = 0;
	while (offset < txt.length()) {
		newblocknumber();
		tempblk.nbr(Lastqueued);
		tempblk.text(txt.substr(offset, Bufferlength));
		offset += Bufferlength;
		TxBlocks.push_back(tempblk);
		Blocks2Send++;
	}
}

void arq::sendblocks()
{
	int framecount = 0;
	cTxtBlk tempblk;
	
	if (TxMissing.empty() == false) {
		list<cTxtBlk>::iterator p = TxMissing.begin();
		while (p != TxMissing.end()) {
			textFrame(*p);
			p++;
			framecount++;
		}
	}
	if (!TxBlocks.empty()) {
		while (TxBlocks.empty() == false && framecount < maxheaders) {
			tempblk = TxBlocks.front();
			if ((tempblk.nbr() + 2)%MAXCOUNT == UrGoodHeader)
				break;
			TxBlocks.pop_front();
			TxMissing.push_back(tempblk);
			textFrame(tempblk);
			LastHeader = tempblk.nbr();
			framecount++;
		}
	}
	if (TxMissing.empty() && TxBlocks.empty())
		Blocks2Send = 0;
	pollFrame();
	LinkState = WAITING;
}
	
void arq::connect(string callsign)
{
	UrCall = callsign;
	LinkState = CONNECTING;
	retrytime = 0; // force a connect attempt
}

void arq::disconnect()
{
	LinkState = DISCONNECTING;
	retrytime = 0; // force a disconnect attempt
}

void arq::abort()
{
	LinkState = ABORTING;
}


void arq::sendUnproto (string txt)
{
	TxTextQueue.erase();
	beaconFrame(txt);
}

void arq::sendPlainText( string txt )
{
	TxPlainTextQueue.append(txt);
}

void arq::transmitdata()
{
	if (TxPlainTextQueue.empty() == false) {
		sendfnc(TxPlainTextQueue);
		TxPlainTextQueue.clear();
		rx2txdelay = TxDelay / ARQLOOPTIME;
		Sending = 0x80;
	} else
	if (TxTextQueue.empty() == false) {
		sendfnc(TxTextQueue);
		TxTextQueue.clear();
		rx2txdelay = TxDelay / ARQLOOPTIME;
		Sending = 0x80;
	}
}

//---------------------------------------------------------------------

void arqloop(void *who)
{
	arq *me = (arq *)who;
	char c;
		
	if (me->getc(c)) {
		me->rcvChar(c);
		while (me->getc(c))
			me->rcvChar(c);
		me->rx2txdelay = me->TxDelay / ARQLOOPTIME;
	} 
	if (me->Sending == 0) { // not transmitting
		if (me->rx2txdelay > 0)
			me->rx2txdelay--;
		else {
			switch (me->LinkState) {
				case  ABORTING :
					me->retrytime = me->RetryTime / ARQLOOPTIME;
					if (me->retries--) {
						me->TxTextQueue.clear();
						me->reset();
						me->abortFrame();
						me->transmitdata();
					}
					break;
				case  CONNECTING :
					if (me->retrytime-- == 0) {
						me->retrytime = me->RetryTime / ARQLOOPTIME;
						if (me->retries--) {
							me->TxTextQueue.clear();
							me->connectFrame();
							me->transmitdata();
						}
					}
					break;
				case  DISCONNECTING :
					if (me->retrytime-- == 0) {
						me->retrytime = me->RetryTime / ARQLOOPTIME;
						if (me->retries--) {
							me->TxTextQueue.clear();
							me->disconnectFrame();
							me->transmitdata();
						}
					}
					break;
				case  WAITING :
					if (me->retrytime-- == 0) {
						me->retrytime = me->RetryTime / ARQLOOPTIME;
						if (me->retries--) {
							me->TxTextQueue.clear();
							me->pollFrame();
							me->transmitdata();
						}
					}
					break;
				case WAITFORACK :
					if (me->retrytime-- == 0) {
						me->retrytime = me->RetryTime / ARQLOOPTIME;
						if (me->retries--) {
							me->TxTextQueue.clear();
							me->ackFrame();
							me->transmitdata();
						}
					}
					break;
				
				case CONNECTED :
					me->timeout--;
				default :
					if ( (me->TxMissing.empty() == false) || 
								(me->TxBlocks.empty() == false))
							me->sendblocks();
					if (me->TxTextQueue.empty() == false ||
						me->TxPlainTextQueue.empty() == false ) {
						me->transmitdata();
					}
					break;
			}
			if (me->retries == 0 || me->timeout == 0) {
				me->LinkState = DOWN;
				me->retries = me->Retries;
				me->retrytime = 0;
				me->timeout = me->Timeout / ARQLOOPTIME;
			}
		}
	}
	if (me->arqstop)
		return;
	Fl::repeat_timeout(	ARQLOOPTIME/1000.0, arqloop, me);
}


void arq::start_arq()
{
	Fl::add_timeout(1.0, arqloop, this);
}

//---------------------------------------------------------------------
